// DFT.CPP Implementation of DFT
// Copyright 2006 Chad Berchek

/* Revisions:

	-2.0 -setNewData function always invalidates results
	     -calculation uses Hanning window
	     -added comments
	     -improved efficiency of calculation
	     -sampling rate must be set in constructor
	     -removed ideal and dual DFT functions
	     -made invalidate function public
	     -changed names from getDualDftAmplitude to getAmplitude; likewise for phase
	     -added getNumSamplesForDFT function to find good frame sizes (number of samples)
	     

	-1.0 first version
	
*/

#include <cmath>
#include "wavetypes.h"
#include "dft.h"
#include "miscmath.h"
using namespace std;

// Constructor for DFTobject. Set some default values.
DFTobject::DFTobject(const samplingRateT setSamplingRate, const sampleT * const dataPtr) : sampleData(dataPtr)
{
	frequency = 0;
	len = 0;
	samplingRate = setSamplingRate;
	invalidate();
}

// Set the frequency that will be measured
void DFTobject::setFreq(frequencyT newFreq) {
	if (newFreq != frequency) {
		frequency = newFreq;
		invalidate();
	}
}

// Set number of samples in sample array
void DFTobject::setLen(sampleCountT n) {
	if (n != len) {
		len = n;
		invalidate();
	}
}

// Set sampling rate
void DFTobject::setSamplingRate(samplingRateT newSamplingRate) {
	if (newSamplingRate != samplingRate) {
		samplingRate = newSamplingRate;
		invalidate();
	}
}

// Set frequency to measure and number of samples in sample array at once
void DFTobject::setFreqAndLen(frequencyT freq, sampleCountT n) {
	if (freq != frequency || n != len) {
		frequency = freq;
		len = n;
		invalidate();
	}
}

// Change sample array
void DFTobject::setNewData(const sampleT * const newData) {
	// Always invalidate
	sampleData = newData;
	invalidate();
}

// Mark current calculated results as invalid
void DFTobject::invalidate() {
	raw.isCurrent = false;
}

// Get measured amplitude of signal
amplitudeT DFTobject::getAmplitude() {
	return 2. * getRawAmplitude(); // Multiply by 2 to compensate for windowing function
}

// Get measured phase of signal
phaseT DFTobject::getPhase() {
	return getRawPhase(); // Phase is not compensated, just use raw phase
}

// Get uncompensated amplitude measurement
amplitudeT DFTobject::getRawAmplitude() {
	if (!raw.isCurrent) Dft();
	return raw.amplitude;
}

// Get uncompensated phase measurement
phaseT DFTobject::getRawPhase() {
	if (!raw.isCurrent) Dft();
	return raw.phase;
}

// Perform DFT on the actual samples. This result is not compensated.
void DFTobject::Dft() {
	if (sampleData != 0) {
		double cosPart = 0, sinPart = 0;
		double arg;
		double hann;
		const double argConstant = 2. * M_PI * frequency / samplingRate; // Compute constant part of arg outside of loop
		const double hannConstant = 2. * M_PI / len; // Compute constant part of hann outside of loop
		for (sampleCountT k = 0; k < len; k++) {
			arg = argConstant * k;  // compute argument of sin and cos functions
			hann = 0.5 * (1 - cos(k * hannConstant)); // compute Hanning window
			sinPart += double(sampleData[k]) * sin(arg) * hann;
			cosPart += double(sampleData[k]) * cos(arg) * hann;
		}
		raw.amplitude = 2. * sqrt( sinPart*sinPart + cosPart*cosPart ) / len;
		if (sinPart == 0 && cosPart == 0) raw.phase = 0;
		else raw.phase = atan2(cosPart, sinPart);
	}
	else {
		raw.amplitude = 0;
		raw.phase = 0;
	}

	raw.isCurrent = true;
}

// Calculate what fraction of a whole cycle is given compared to the nearest whole number of cycles
// This is a support function for getNumSamplesForDFT
double getFractionOfCycle(const double cycles) {
	double fracCycle = fpart(cycles);
	double from1 = 1. - fracCycle;
	double x;

	x = (fracCycle <= from1) ? fracCycle : from1;
	return x;
}

// getNumSamples - get a suitable number of samples needed for an accurate DFT
// freq = frequency that will be measured
// samplingRate = sampling rate
// toneDuration = desired duration of samples used in DFT
// maxFractionalCycle = maximum fraction of a cycle of the input frequency that will be allowed
// minNumCycles = the minimum number of cycles of the input frequency that will be allowed
// maxExtraSamples = the maximum number of extra samples that will be added to the desired tone duration for accuracy
unsigned int getNumSamplesForDFT(const double freq, const unsigned int samplingRate, const double toneDuration,
	const double maxFractionalCycle, const unsigned int minNumCycles, const unsigned int maxExtraSamples) {

	if ((freq > samplingRate / 2) || (freq <= 0)) return 0; // Avoid invalid results if freq out of range

	double errorFreq = 2. * freq;
	if (errorFreq > samplingRate / 2.) errorFreq = samplingRate - errorFreq;
	if (errorFreq == 0.) return (unsigned int)(toneDuration * samplingRate);

	double eqFreq = errorFreq / 2.;  // Equivalent frequency - frequency
               	// which has the same error frequency as the actual frequency
	double numCycles = ceil( toneDuration * eqFreq );
	if (numCycles < minNumCycles) numCycles = minNumCycles;
	double realNumSamples;
	double bestFraction = 1.;
	double bestNumSamples = numCycles / eqFreq * samplingRate;
	double fractionalCycle;
	double iNumSamples;

	double extraCycles = 0;
	do {
		realNumSamples = double(numCycles + extraCycles) / eqFreq * samplingRate;
		iNumSamples = fround(realNumSamples);
      fractionalCycle = getFractionOfCycle(double(iNumSamples) / samplingRate * eqFreq);
		if (fractionalCycle < bestFraction) {
			bestFraction = fractionalCycle;
			bestNumSamples = realNumSamples;
		}
		if (bestFraction <= maxFractionalCycle) break;
		extraCycles += 1.;
	} while ((extraCycles / eqFreq * samplingRate) < maxExtraSamples);

	return (unsigned int)fround(bestNumSamples);
}
