// Harmonalyzer main program.
// Contains harmonic analysis, wave file processing, and results output.
// Copyright 2006 Chad Berchek

/* Revisions:

	-1.0 first version
*/

#include <cstdlib>
#include <cstdio>
#include "wave.h"
#include "wavetypes.h"
#include "dft.h"
#include "hasettings.h"
#include "userinput.h"
#include "miscmath.h"
#include "kbinput.h"
#include "pitch.h"
using namespace std;

// Get the number of samples needed for an accurate DFT at the given frequency. Basically this function just
// converts the data from an haSettings struct to a call to getNumSamplesForDFT
unsigned int getNumSamplesForFreq(const double freq, const haSettings &haset) {
	return getNumSamplesForDFT(freq, haset.sampleRate, haset.toneDuration, haset.maxFractionalCycle, haset.minNumCycles,
		sampleCountT(haset.maxExtraTime * haset.sampleRate));
}

// Finds the maximum number of samples needed to perform an accurate DFT for the given harmonics and fundamental.
// fundamentalFreq = the fundamental frequency
// haset = harmonic analyzer settings structure
unsigned int findNumSamplesForDFT(const double fundamentalFreq, const haSettings &haset) {
	unsigned int maxNumSamples = 0;
	
	for (unsigned int fundamentalMult = 1; fundamentalMult <= haset.numHarmonics + 1; fundamentalMult++) {
		unsigned int numSamplesForHarmonic = getNumSamplesForFreq(fundamentalFreq * fundamentalMult, haset);
		if (numSamplesForHarmonic > maxNumSamples) maxNumSamples = numSamplesForHarmonic;
	}
	
	return maxNumSamples;
}

// Checks to see if the current position in the wave file is in a break of silence. If so, the wave
// file position is left at the end of the set of samples checked for silence, as determined by
// haset.breakLengthThreshold, and the function returns true. If it is determined that there is
// not a break of silence, the wave file is rewound to the position where it was when the function
// started (i.e., the wave file state is unchanged after the function call) and the function returns
// false.
bool isBreak(WaveFile &inWave, const haSettings &haset) {
	sampleCountT breakLengthThresholdS = (sampleCountT)(haset.breakLengthThreshold * haset.sampleRate);
	for (unsigned int samplesChecked = 1; samplesChecked <= breakLengthThresholdS; samplesChecked++) {
		short sample;
		bool success = inWave.ReadSample(sample);
		if (!success) break;
		if (abs(sample) > haset.silenceThreshold) {
			inWave.Rewind(samplesChecked);
			return false;
		}
	}
	return true;
}

// analyzeHarmonics measures the harmonics in the signal
// samples = array of wave samples
// SAMPLES_LEN = number of samples in "samples" array
// harmonicAmplitude = array to hold the amplitude of each harmonic. harmonicAmplitude[0] = fundamental amplitude
// numHarmonics = number of harmonics to analyze. 0 = fundamental only. There must be numHarmonics+1 entries in the
//    harmonicAmplitude array.
// fundamentalFreq = the fundamental frequency
// harmonicAmplitude[h] is -1 if the frequency of harmonic h is above half the sampling rate
void analyzeHarmonics(const double * const samples, const unsigned int SAMPLES_LEN, double * const harmonicAmplitude,
	const double fundamentalFreq, const haSettings &haset) {

	DFTobject dft(haset.sampleRate, samples);

	for (unsigned int harmonic = 0; harmonic <= haset.numHarmonics; harmonic++) {
		double currentFreq = fundamentalFreq * (harmonic + 1);
		if (currentFreq > haset.sampleRate / 2) harmonicAmplitude[harmonic] = -1;
		else {
			unsigned int numSamplesForHarmonic = getNumSamplesForFreq(currentFreq, haset);
			if (numSamplesForHarmonic > SAMPLES_LEN) numSamplesForHarmonic = SAMPLES_LEN;
			dft.setFreqAndLen(currentFreq, numSamplesForHarmonic);
			harmonicAmplitude[harmonic] = dft.getAmplitude();
		}
	}
}

// Process the whole wave file. This function detects tones then calls functions
// that perform pitch detection and harmonic analysis. Then the output is written
// to a file.
void analyzeWaveFile(WaveFile &inWave, const haSettings &haset, FILE *outFile) {
	const unsigned int totalPitchDetectSamples = haset.FFT_LEN + haset.osampOffset;
	double *pitchDetectSamples = new double[totalPitchDetectSamples];
	bool readSuccess;
	short sample;
	double *harmonicAmpl = new double[haset.numHarmonics + 1];    // Array for harmonic amplitudes
	double *harmonicAmplSum = new double[haset.numHarmonics + 1]; // Plus 1 because it needs enough space
																						 // for the fundamental and harmonics
	// Each iteration of this loop should process one tone. Tones are separated by breaks of silence.
	do {
		// Skip past silence
		do {
			readSuccess = inWave.ReadSample(sample);
		} while (sample <= haset.silenceThreshold && readSuccess);
		if (!readSuccess) break;

		// Skip attack of note
		readSuccess = inWave.SeekFwd(sampleCountT(haset.attackSkipDuration * haset.sampleRate));
		if (!readSuccess) break;

		// Variables for the series of analysis frames to be done
		bool breakDetected; // Break between tones detected?
		unsigned int numOfFrames = 0; // Number of frames done for this tone
		for (unsigned int i = 0; i <= haset.numHarmonics; i++) harmonicAmplSum[i] = 0; // Zero array
		double maxFundAmpl = 0; // Maximum amplitude of fundamental
		double fundFreqSum = 0; // Sum of fundamental frequencies, for determining average freq.

		// Do as many frames as the duration of the tone allows. Average results.
		// Each iteration of this loop should process one set of DFTs for the harmonics.
		do {
			// Read in samples for pitch detection
			for (unsigned int i = 0; i < totalPitchDetectSamples; i++) {
				readSuccess = inWave.ReadSample(sample);
				pitchDetectSamples[i] = sample;
				if (abs(sample) <= haset.silenceThreshold) { // Check for end of tone
					breakDetected = isBreak(inWave, haset);
					if (breakDetected) break;
				}
			}
			if (breakDetected || !readSuccess) break;  // If end of tone reached, do not process this frame

			// Do pitch detection
			foundFund ff = findFundamental(pitchDetectSamples, totalPitchDetectSamples,
				haset.FFT_LEN, haset.osampOffset, inWave.GetSampleRate(), haset.peakDetectThreshold, haset.numPeaks);
			if (ff.errorCode == 2) {
				printf("Error: no peaks found. Peak detect threshold may be too high.\n");
				continue;
			}
			else if (ff.errorCode != 0) { // check for other errors
				printf("Error: fundamental frequency detection error. Please report this error.\n");
				continue;
			}
			double fundFreq = ff.frequency;
			
			// Print a message so the user knows what's going on.
			printf("Detected tone at %.2f Hz\n", fundFreq);

			// Find number of samples needed for the DFTs of the harmonics. Allocate the sample array.
			unsigned int numSamplesForDFT = findNumSamplesForDFT(fundFreq, haset);
			double *samplesForDFT = new double[numSamplesForDFT];

			// Copy samples from frequency measurement to samplesForDFT array
			unsigned int copySamples = min(totalPitchDetectSamples, numSamplesForDFT);
			for (unsigned int i = 0; i < copySamples; i++) samplesForDFT[i] = pitchDetectSamples[i];

			// Read samples for harmonic analysis
			for (unsigned int i = copySamples; i < numSamplesForDFT; i++) {
				readSuccess = inWave.ReadSample(sample);
				samplesForDFT[i] = sample;
				if (abs(sample) <= haset.silenceThreshold) { // Check for end of tone
					breakDetected = isBreak(inWave, haset);
					if (breakDetected) break;
				}
			}
			if (!breakDetected && readSuccess) {  // If after reading in the samples the input position is
			                                      // in the middle of a break, don't use these samples.
				// Analyze harmonics
				analyzeHarmonics(samplesForDFT, numSamplesForDFT, harmonicAmpl, fundFreq, haset);

				// Add results
				for (unsigned int i = 0; i <= haset.numHarmonics; i++) harmonicAmplSum[i] += harmonicAmpl[i];
				fundFreqSum += fundFreq;
				numOfFrames++; // Increment count of the number of frames done

				// Detect max fundamental
				if (harmonicAmpl[0] > maxFundAmpl) maxFundAmpl = harmonicAmpl[0];
			}

			delete[] samplesForDFT;
		} while (!breakDetected && readSuccess);

		// Write results to output file. Format: fundFreq, maxFundAmpl, harmonic1AvgAmpl, harmonic2AvgAmpl, ...
		if (numOfFrames > 0) {
			fprintf(outFile, "%.3f, %.1f", fundFreqSum / numOfFrames, maxFundAmpl);
			for (unsigned int harmonic = 0; harmonic <= haset.numHarmonics; harmonic++) {
				if (harmonicAmplSum[harmonic] < 0) fprintf(outFile, ", "); // Error: harmonic freq. greater than sampling rate / 2
            else fprintf(outFile, ", %.1f", double(harmonicAmplSum[harmonic] / numOfFrames)); // Result OK
			}
			fprintf(outFile, "\n");
		}
		else {
			fprintf(outFile, "Tone too short to measure\n");
		}

	} while (readSuccess);

	delete[] pitchDetectSamples;
	delete[] harmonicAmpl;
	delete[] harmonicAmplSum;
}

// Write the column headers to the output file.
// Format: Frequency, Max Fundamental Amplitude, Harmonic 0, Harmonic 1, ...
void writeOutFileHeader(FILE * const outFile, const haSettings &haset) {
	fprintf(outFile, "Frequency, Max Fundamental Amplitude");
	for (unsigned int harmonic = 0; harmonic <= haset.numHarmonics; harmonic++) {
		fprintf(outFile, ", Harmonic %u", harmonic);
	}
	fprintf(outFile, "\n");
}


// Check that input wave file is mono 16-bit.
// Display error message if there's a problem.
// Returns true if input wave file is OK, otherwise false.
bool checkWaveFormat(WaveFile &inWave) {
	bool OK = true;
	if (inWave.GetNumChannels() != 1) {
		printf("\nError: the input wave file must be mono (1 channel).\n");
		OK = false;
	}
	if (inWave.GetBitsPerChannel() != 16) {
		printf("\nError: the input wave file must have 16-bit samples.\n");
		OK = false;
	}
	return OK;
}

int main() {
	WaveFile inWave;
	FILE *outFile;
	haSettings haset;

	bool openFilesOK = openFiles(inWave, outFile);
	if (!openFilesOK) exit(0);

	bool inputWaveFormatOK = checkWaveFormat(inWave);
	if (!inputWaveFormatOK) {
		printf("Press <Enter> to quit.\n");
		kbReadEnter();
		exit(1);
	}

	enterUserSettings(haset, inWave.GetSampleRate());

	writeOutFileHeader(outFile, haset);
	analyzeWaveFile(inWave, haset, outFile);
	
	fclose(outFile);
	
	printf("Press <Enter> when done viewing output.\n");
	kbReadEnter();
	
	return 0;
}
