// Pitch.cpp - contains functions for finding pitch / detecting frequency
// Copyright 2006 Chad Berchek

/* Revisions:
	1.0 - first version
*/

#include <cstdlib>
#include "miscmath.h"
#include "fourier.h"
#include "pitch.h"
using namespace std;

// Find a peak in the amplitude spectrum, starting from the given
// bin and moving right. A peak is defined as a bin whose magnitude
// is greater than or equal to the threshold and greater than the
// magnitude of the neighboring bins.
// Parameters:
// ampl = amplitude spectrum
// startBin = index of first bin to consider (index starts at 0)
// NUM_BINS = number of bins to consider = number of numbers in ampl array
// threshold = lowest value that can be detected as a peak
foundBin findPeak(const double * const ampl, const unsigned int startBin,
	const unsigned int NUM_BINS, const double thresholdAmplitude) {

	unsigned int bin;
	unsigned int peakBin = 0;
	bool foundPeak = false;
	for (bin = startBin; bin < NUM_BINS; bin++) {
		if (ampl[bin] >= thresholdAmplitude) {
			foundPeak = true; // Assume for a moment that this is a peak
			if (bin > 0) { // ensure bin - 1 is a valid index
            // check that this bin is bigger than the bin to the left
				if (ampl[bin] < ampl[bin - 1]) foundPeak = false;
			}
			if (bin < NUM_BINS - 1) { // ensure bin + 1 is a valid index
			   // check that this bin is bigger than the bin to the right
			   if (ampl[bin] < ampl[bin + 1]) foundPeak = false;
			}
			if (foundPeak) {
				peakBin = bin;
				break; // if peak found, quit searching
			}
		}
	}

	foundBin pk;
	pk.isValid = foundPeak;
	pk.binIndex = peakBin;
	return pk;
}

// Find the bin containing the fundamental frequency. To accomplish this, the peaks of the
// spectrum are examined. The difference between the peaks, which are assumed to be harmonics,
// is calculated and taken as the fundamental. The DC component (bin 0) is ignored when searching
// for peaks.
// Parameters:
// ampl = amplitude spectrum
// NUM_BINS = number of bins = number of numbers in ampl array
// avgMultThreshold = average multiplier threshold. The threshold value will be the average amplitude
//    times the average multiplier threshold.
// numPeaks = number of peaks to consider
foundBin findFundBin(const double * const ampl, const unsigned int NUM_BINS,
	const double avgMultThreshold, const unsigned int numPeaks) {

	foundBin fundBin; // Fundamental bin
	fundBin.binIndex = 0;
	unsigned int fundBinIndex = 0;
	unsigned int bin;

	// Find average amplitude
	double amplSum = 0;
	for (bin = 0; bin < NUM_BINS; bin++) amplSum += ampl[bin];
	double avgAmpl = amplSum / NUM_BINS;

	// Find actual amplitude threshold for peak detection
	double thresholdAmplitude = avgAmpl * avgMultThreshold;

	// Allocate array to hold the peaks
	unsigned int *peaks = new unsigned int[numPeaks];

	// Find the peaks
	unsigned int peakNumber;
	unsigned int lastFoundPeak = 0; // Peak search will start at lastFoundPeak + 1, which means
											  // the first bin considered will be 1. This ignores the DC component.
	for (peakNumber = 0; peakNumber < numPeaks; peakNumber++) {
		foundBin pk = findPeak(ampl, lastFoundPeak + 1, NUM_BINS, thresholdAmplitude);
		if (pk.isValid) {
			peaks[peakNumber] = pk.binIndex;
			lastFoundPeak = pk.binIndex;
		}
		else break;
	}
	const unsigned int numPeaksFound = peakNumber;

	if (numPeaksFound > 0) { // Proceed with processing only if peaks found
		int *diffBetweenPeaks = new int[numPeaksFound];

		// Calculate difference, in bins, between peaks
		diffBetweenPeaks[0] = peaks[0];
		for (peakNumber = 1; peakNumber < numPeaksFound; peakNumber++) {
			diffBetweenPeaks[peakNumber] = peaks[peakNumber] - peaks[peakNumber - 1];
		}

		// Find the number of times each difference appears in diffBetweenPeaks
		int *hist = new int[numPeaksFound]; // histogram
		for (peakNumber = 0; peakNumber < numPeaksFound; peakNumber++) {
			hist[peakNumber] = 0; // Start hit count at 0
			for (unsigned int i = 0; i < numPeaksFound; i++) {
				// If diffs are within 1 of each other, increment the hit count for that diff
				if (isWithin(diffBetweenPeaks[peakNumber], diffBetweenPeaks[i], 1)) hist[peakNumber]++;
				// Also check if half of the one diff is within .5 of the other diff in case of a missing harmonic between two peaks
				else if (isWithinD(diffBetweenPeaks[peakNumber], double(diffBetweenPeaks[i]) / 2., 0.501)) hist[peakNumber]++;
			}
		}

		// Determine which difference has the most hits and take it as a bin near the fundamental
		unsigned int searchCenterBin = 0; // Bin that will be the center of the final search for the fundamental
		int maxHits = -1; // max hits found so far
		for (unsigned int i = 0; i < numPeaksFound; i++) {
			if (hist[i] > maxHits) {
				maxHits = hist[i];
				searchCenterBin = diffBetweenPeaks[i];
			}
		}

		// Search the 3 bins centered around searchCenterBin for the maximum amplitude
		double maxAmpl = -1; // Maximum amplitude found so far
		for (int bin = searchCenterBin - 1; bin <= searchCenterBin + 1; bin++) {
			if (bin >= 0 && bin < NUM_BINS) {
				if (ampl[bin] > maxAmpl) {
					maxAmpl = ampl[bin];
					fundBin.binIndex = bin;
				}
			}
		}

		fundBin.isValid = true;

		delete[] hist;
		delete[] diffBetweenPeaks;
	}
	else fundBin.isValid = false;


	delete[] peaks;
	return fundBin;
}

// applyHanningWindow Notes:
// inSamples = input samples
// outSamples = output samples
// inSamples can = outSamples (do operation in-place)
void applyHanningWindow(const double * const inSamples, double * const outSamples, const unsigned int NUM_SAMPLES) {
	for (size_t i = 0; i < NUM_SAMPLES; i++) {
		outSamples[i] = inSamples[i] * 0.5 * (1 - cos(i * M_2PI / NUM_SAMPLES));
	}
}

// Get amplitude of FFT bins.
// real, imag = output values from FFT
// ampl = array to hold the amplitude for each bin
// NUM_BINS = number of bins in real, imag, and ampl
void computeAmplitudes(const double * const real, const double * const imag, double * const ampl, const unsigned int NUM_BINS) {
	for (unsigned int bin = 0; bin < NUM_BINS; bin++) ampl[bin] = sqrt( real[bin] * real[bin] + imag[bin] * imag[bin] );
}

// findTrueFreq description:
// returns the true frequency of the bin
// binIndex = the index in the FFT results of the bin to process
// phase1 = the phase of the bin in frame 1
// phase2 = the phase of the bin in frame 2
// osampOffset = the number of samples elapsed between the start of frame 1 and the start of frame 2 - that is, if frame 1 began
//    at sample 0, then frame 2 began at sample osampOffset
// FFT_LEN the length of the FFT frame in samples
// SAMPLE_RATE the sampling rate in Hz
double findTrueFreq(const unsigned int binIndex, const double phase1, const double phase2, const unsigned int osampOffset,
	const unsigned int FFT_LEN, const unsigned int SAMPLE_RATE) {

	double fb = Index_to_frequency(FFT_LEN, binIndex) * SAMPLE_RATE;  // bin frequency
	double phim = mapToPi(phase2 - phase1);  // phi m - measured phase difference
	double phimb = mapToPi(M_2PI * osampOffset * fb / SAMPLE_RATE); // phi mb - expected measured phase difference if input frequency = bin frequency
	while ((phim - phimb) < -M_PI) phimb -= M_2PI;  // difference in phase should be no greater than pi
	while ((phim - phimb) >  M_PI) phimb += M_2PI;
	double realFreq = double(SAMPLE_RATE) / (M_2PI * osampOffset) * (phim - phimb) + fb;
	return realFreq;
}

// findFundamental description:
// findFundamental returns the true frequency of the fundamental frequency present
// samples = pointer to the digital waveform samples
// SAMPLES_LEN = the number of samples available in the samples array
// FFT_LEN = the length of FFT to use. recommended not less than 1024. must be a power of 2.
// osampOffset = the number of samples between the two frames. recommended about half of FFT_LEN
// SAMPLE_RATE = the sampling rate in Hz
// peakDetectThreshold = threshold for detecting peaks in the spectrum. The actual threshold will be the
//    average amplitude of the spectrum times peakDetectThreshold
// numPeaks = number of peaks to consider when searching for fundamental
// Error codes:
//     0 = no error
//     1 = not enough samples in sample array for given FFT length and oversampling offset
//     2 = no peaks found
foundFund findFundamental(const double * const samples, const unsigned int SAMPLES_LEN, const unsigned int FFT_LEN,
	const unsigned int osampOffset, const unsigned int SAMPLE_RATE, const unsigned int peakDetectThreshold,
	const unsigned int numPeaks) {

	const unsigned int FFT_LEN_2 = FFT_LEN / 2;
	foundFund retval;
	retval.errorCode = 0;

	if (SAMPLES_LEN < FFT_LEN + osampOffset) {
		retval.errorCode = 1;
	}
	else {
		double *realIn = new double[FFT_LEN];
		double *realOut = new double[FFT_LEN];
		double *imagOut = new double[FFT_LEN];
		double *ampl = new double[FFT_LEN_2];

		applyHanningWindow(samples, realIn, FFT_LEN);  // apply Hanning window to samples to be used in FFT
		fft_double(FFT_LEN, 0, realIn, NULL, realOut, imagOut);  // perform the first FFT
		computeAmplitudes(realOut, imagOut, ampl, FFT_LEN_2);
		foundBin fundBin = findFundBin(ampl, FFT_LEN_2, peakDetectThreshold, numPeaks);
		const unsigned int fundBinIndex = fundBin.binIndex;

		if (!fundBin.isValid) {
			retval.errorCode = 2;
		}
		else {
			double phase1 = -safeAtan2(imagOut[fundBinIndex], realOut[fundBinIndex]);

			applyHanningWindow(samples + osampOffset, realIn, FFT_LEN);  // apply Hanning window to samples to be used in FFT
			fft_double(FFT_LEN, 0, realIn, NULL, realOut, imagOut);  // perform the second FFT
			double phase2 = -safeAtan2(imagOut[fundBinIndex], realOut[fundBinIndex]);

			double trueFundamental = findTrueFreq(fundBinIndex, phase1, phase2, osampOffset, FFT_LEN, SAMPLE_RATE);
			retval.frequency = trueFundamental;
		}

		delete[] realIn;
		delete[] realOut;
		delete[] imagOut;
		delete[] ampl;
	}

	return retval;
}
