// Userinput.cpp
// Contains openFiles function for opening the wave input and text output files
// and enterUserSettings function for getting the user's settings.

// Copyright 2006 Chad Berchek

// *** Un-comment the line below to build for Windows platforms.
//#define USE_MS_WIN_DIALOGS

#define _WIN32_WINNT 0x0400 // Keep this line if building for Windows 95/98

#include <cstdio>
#include "kbinput.h"
#include "wavetypes.h"
#include "hasettings.h"
#include "userinput.h"
#include "wave.h"
using namespace std;

const size_t FILE_NAME_BUF_SIZE = 1024;

// User settings min max and default values:
const unsigned int MIN_NUM_HARMONICS = 0;
const unsigned int MAX_NUM_HARMONICS = 50;
const unsigned int DEF_NUM_HARMONICS = 10;
const double MIN_TONE_DURATION = 0.05;
const double MAX_TONE_DURATION = 30;
const double DEF_TONE_DURATION = 0.3;
const unsigned int MIN_OSAMP_OFFSET = 1;
const unsigned int DEF_OSAMP_OFFSET = 1024;
const double MIN_BREAK_LENGTH = 0.1;
const double MAX_BREAK_LENGTH = 10;
const double DEF_BREAK_LENGTH = 0.4;
const unsigned int MIN_PEAK_THRESHOLD = 1;
const unsigned int MAX_PEAK_THRESHOLD = 150;
const unsigned int DEF_PEAK_THRESHOLD = 40;
const unsigned int MIN_SILENCE_THRESHOLD = 1;
const unsigned int MAX_SILENCE_THRESHOLD = 10000;
const unsigned int DEF_SILENCE_THRESHOLD = 3200;
const unsigned int MIN_MIN_NUM_CYCLES = 1;
const unsigned int MAX_MIN_NUM_CYCLES = 100;
const unsigned int DEF_MIN_NUM_CYCLES = 5;
const double MIN_MAX_EXTRA_TIME = 0;
const double MAX_MAX_EXTRA_TIME = 1;
const double DEF_MAX_EXTRA_TIME = 0.04;
const double MIN_MAX_FRAC_CYCLE = 0.0001;
const double MAX_MAX_FRAC_CYCLE = 1;
const double DEF_MAX_FRAC_CYCLE = 0.01;
const double MIN_ATTACK_SKIP = 0;
const double MAX_ATTACK_SKIP = 10;
const double DEF_ATTACK_SKIP = 0.2;
const unsigned int MIN_NUM_PEAKS = 1;
const unsigned int MAX_NUM_PEAKS = 20;
const unsigned int DEF_NUM_PEAKS = 5;

// Open files using Windows dialogs or text file-name entry
// Returns false if opening files is canceled, true if opening files succeeds.
#ifdef USE_MS_WIN_DIALOGS
	#include <windows.h>
	bool openFiles(WaveFile &waveFile, FILE* &outputFile) {
		bool openSuccess;
		OPENFILENAME ofn = { 0 };
		char fileNameBuf[FILE_NAME_BUF_SIZE];

		// Initialize ofn structure
		ofn.lStructSize = sizeof(ofn);
		ofn.lpstrFile = fileNameBuf;
		ofn.nMaxFile = FILE_NAME_BUF_SIZE;

		//////// Open wave file for input //////////
		fileNameBuf[0] = 0;
		const char *wavFilter = "Wave Files (*.wav)\0*.wav\0All Files (*.*)\0*\0";
		ofn.lpstrFilter = wavFilter;
		ofn.Flags = OFN_FILEMUSTEXIST;
		ofn.lpstrTitle = "Select the recorded sample";
		do {
			BOOL success = GetOpenFileName(&ofn);
			if (!success) return false;
			openSuccess = waveFile.OpenRead(fileNameBuf);
			if (!openSuccess) {
				MessageBox(0, "The selected file could not be opened as a wave file.", "Error", MB_OK | MB_ICONEXCLAMATION);
			}
		} while (!openSuccess);

		// Copy input wave filename to string (without path)
		char *waveFileName = new char[FILE_NAME_BUF_SIZE];
		strcpy(waveFileName, fileNameBuf + ofn.nFileOffset);
		// Find position of '.' before filename extension
		long waveFileNameFullLen = strlen(fileNameBuf);
		long waveFileNameExtnPos;
		if ((ofn.nFileExtension == 0) || (ofn.nFileExtension == waveFileNameFullLen)) {
			waveFileNameExtnPos = waveFileNameFullLen - ofn.nFileOffset;
		}
		else waveFileNameExtnPos = ofn.nFileExtension - 1 - ofn.nFileOffset;  // -1 takes care of '.'
		// Initialize filename to wave file name minus extension
		strcpy(fileNameBuf, waveFileName);
		fileNameBuf[waveFileNameExtnPos] = 0;  // truncate extension
		delete[] waveFileName;

		//////// Open text file for output //////////
		const char *txtFilter = "Text Files (*.txt)\0*.txt\0All Files (*.*)\0*\0";
		ofn.lpstrFilter = txtFilter;
		ofn.lpstrTitle = "Enter a filename for the output";
		ofn.lpstrDefExt = "txt";
		ofn.Flags = OFN_NOREADONLYRETURN | OFN_OVERWRITEPROMPT;
		do {
			BOOL success = GetSaveFileName(&ofn);
			if (!success) return false;
			outputFile = fopen(fileNameBuf, "wt");
			if (!outputFile) {
				MessageBox(0, "The selected file could not be opened for output.", "Error", MB_OK | MB_ICONEXCLAMATION);
			}
		} while (!outputFile);
		
		return true;
	}
#else
	bool openFiles(WaveFile &waveFile, FILE* &outputFile) {
		char buf[FILE_NAME_BUF_SIZE];

		printf("\n(Note: Enter a blank filename to quit the program.)");

		// Open wave file for input
		do {
			printf("\nEnter recorded wave filename: ");
			kbReadString(buf, FILE_NAME_BUF_SIZE);
			if (isEmptyString(buf)) return false;
		} while (!waveFile.OpenRead(buf));

		do {
			printf("\nEnter output text filename: ");
			kbReadString(buf, FILE_NAME_BUF_SIZE);
			if (isEmptyString(buf)) return false;
			outputFile = fopen(buf, "wt");
		} while (!outputFile);

		return true;
	}
#endif

// Convert FFT size code to actual FFT size
// 1=2048 samples, 2=4096, 3=8192, 4=16384
sampleCountT getFFTsize(unsigned int FFTsizeCode) {
	return 1024 * (1 << FFTsizeCode);
}

// Get the user's input for settings in haSettings struct
void enterUserSettings(haSettings &haset, const samplingRateT sampleRate) {
	// Setup FFT size code
	unsigned int FFTsizeCode;
	unsigned int defFFTsizeCode;
	if (sampleRate <= 22050) defFFTsizeCode = 2;
	else defFFTsizeCode = 3;

	// Load defaults
	haset.sampleRate = sampleRate;
	haset.numHarmonics = DEF_NUM_HARMONICS;
	haset.toneDuration = DEF_TONE_DURATION;
	haset.attackSkipDuration = DEF_ATTACK_SKIP;
	haset.FFT_LEN = getFFTsize(defFFTsizeCode);
	haset.osampOffset = DEF_OSAMP_OFFSET;
	haset.breakLengthThreshold = DEF_BREAK_LENGTH;
	haset.peakDetectThreshold = DEF_PEAK_THRESHOLD;
	haset.silenceThreshold = DEF_SILENCE_THRESHOLD;
	haset.minNumCycles = DEF_MIN_NUM_CYCLES;
	haset.maxExtraTime = DEF_MAX_EXTRA_TIME;
	haset.maxFractionalCycle = DEF_MAX_FRAC_CYCLE;
	haset.numPeaks = DEF_NUM_PEAKS;

	// Modify defaults?
	printf("\nDo you wish to modify the default settings? (Type yes or no then press <Enter>) ");
	bool modifyDefaults = kbReadYN();

	if (modifyDefaults) {
		printf("\n\n*** NOTE: For any setting, just press <Enter> to use the default value. ***\n");
		
		printf("\nEnter the number of harmonics to measure (0 = measure only fundamental)");
		printf("\n(Minimum = %u, Maximum = %u, Default = %u): ", MIN_NUM_HARMONICS, MAX_NUM_HARMONICS, DEF_NUM_HARMONICS);
		haset.numHarmonics = (unsigned int)kbReadDouble(MIN_NUM_HARMONICS, MAX_NUM_HARMONICS, DEF_NUM_HARMONICS);
		
		printf("\nEnter the minimum tone duration in seconds.");
		printf("\n(Minimum = %.2f, Maximum = %.1f, Default = %.1f): ", MIN_TONE_DURATION, MAX_TONE_DURATION, DEF_TONE_DURATION);
		haset.toneDuration = kbReadDouble(MIN_TONE_DURATION, MAX_TONE_DURATION, DEF_TONE_DURATION);
		
		printf("\nEnter the length of the note attack in seconds. This length will be skipped");
		printf("\nat the beginning of each tone. (Minimum = %.0f, Maximum = %.0f, Default = %.1f): ", MIN_ATTACK_SKIP, MAX_ATTACK_SKIP, DEF_ATTACK_SKIP);
		haset.attackSkipDuration = kbReadDouble(MIN_ATTACK_SKIP, MAX_ATTACK_SKIP, DEF_ATTACK_SKIP);
		
		printf("\nEnter the peak detection threshold for pitch detection");
		printf("\n(Minimum = %u, Maximum = %u, Default = %u): ", MIN_PEAK_THRESHOLD, MAX_PEAK_THRESHOLD, DEF_PEAK_THRESHOLD);
		haset.peakDetectThreshold = (unsigned int)kbReadDouble(MIN_PEAK_THRESHOLD, MAX_PEAK_THRESHOLD, DEF_PEAK_THRESHOLD);

		// Modify "advanced" settings?
		printf("\nDo you wish to modify the \"advanced\" settings? (yes / no) ");
		bool modifyAdvanced = kbReadYN();
		
		if (modifyAdvanced) {
			printf("\nEnter the length of silence in seconds that will indicate a break between tones");
			printf("\n(Minimum = %.1f, Maximum = %.1f, Default = %.1f): ", MIN_BREAK_LENGTH, MAX_BREAK_LENGTH, DEF_BREAK_LENGTH);
			haset.breakLengthThreshold = kbReadDouble(MIN_BREAK_LENGTH, MAX_BREAK_LENGTH, DEF_BREAK_LENGTH);

			printf("\nSelect the size of the FFT to use for pitch detection.");
			printf("\n  1.  2048");
			printf("\n  2.  4096");
			printf("\n  3.  8192");
			printf("\n  4. 16384");
			printf("\nEnter a number 1-4 (default = %u): ", defFFTsizeCode);
			FFTsizeCode = (unsigned int)kbReadDouble(1, 4.01, defFFTsizeCode);
			haset.FFT_LEN = getFFTsize(FFTsizeCode);

			const unsigned int MAX_OSAMP_OFFSET = haset.FFT_LEN;
			printf("\nEnter the number of samples between the two FFT frames for pitch detection");
			printf("\n(Minimum = %u, Maximum = %u, Default = %u): ", MIN_OSAMP_OFFSET, MAX_OSAMP_OFFSET, DEF_OSAMP_OFFSET);
			haset.osampOffset = sampleCountT(kbReadDouble(MIN_OSAMP_OFFSET, MAX_OSAMP_OFFSET, DEF_OSAMP_OFFSET));

			printf("\nEnter the silence threshold (the maximum sample value that will be considered");
			printf("\nsilent) (Minimum: %u, Maximum: %u, Default: %u): ", MIN_SILENCE_THRESHOLD, MAX_SILENCE_THRESHOLD, DEF_SILENCE_THRESHOLD);
			haset.silenceThreshold = (sampleT)kbReadDouble(MIN_SILENCE_THRESHOLD, MAX_SILENCE_THRESHOLD, DEF_SILENCE_THRESHOLD);

			printf("\nEnter the minimum number of cycles of a frequency to allow");
			printf("\n(Minimum = %u, Maximum = %u, Default = %u): ", MIN_MIN_NUM_CYCLES, MAX_MIN_NUM_CYCLES, DEF_MIN_NUM_CYCLES);
			haset.minNumCycles = (cycleCountT)kbReadDouble(MIN_MIN_NUM_CYCLES, MAX_MIN_NUM_CYCLES, DEF_MIN_NUM_CYCLES);

			printf("\nEnter the maximum extra time for the DFT frame beyond the normal tone duration");
			printf("\n(Minimum = %.0f, Maximum = %.0f, Default = %.2f): ", MIN_MAX_EXTRA_TIME, MAX_MAX_EXTRA_TIME, DEF_MAX_EXTRA_TIME);
			haset.maxExtraTime = kbReadDouble(MIN_MAX_EXTRA_TIME, MAX_MAX_EXTRA_TIME, DEF_MAX_EXTRA_TIME);

			printf("\nEnter the maximum fraction of a cycle to allow in a DFT frame");
			printf("\n(Minimum = %.4f, Maximum = %.0f, Default = %.2f): ", MIN_MAX_FRAC_CYCLE, MAX_MAX_FRAC_CYCLE, DEF_MAX_FRAC_CYCLE);
			haset.maxFractionalCycle = kbReadDouble(MIN_MAX_FRAC_CYCLE, MAX_MAX_FRAC_CYCLE, DEF_MAX_FRAC_CYCLE);

			printf("\nEnter the number of peaks to consider when searching for the fundamental");
			printf("\n(Minimum = %u, Maximum = %u, Default = %u): ", MIN_NUM_PEAKS, MAX_NUM_PEAKS, DEF_NUM_PEAKS);
			haset.numPeaks = (unsigned int)kbReadDouble(MIN_NUM_PEAKS, MAX_NUM_PEAKS, DEF_NUM_PEAKS);
		}
	}
}
